use strict;

use Net::Domain qw(hostfqdn);	# $ENV{HOST} and system hostname are unreliable

my $tempfile_dir = $ENV{HOME} . '/.grid_dist/';

my $ws_timeout_msec = 3 * 60 * 1000;

my $debug = '';

=pod

=head2 Job

=over 1

Class representing distributed grid compute jobs.

=back
=head3 Each Job has properties

=over 2

=item   NAME    (the IP host name)
=item	TIMESTAMP a timestamp uniquely identifying the run of the grid_dist
        script
=item	LOGFILE path on the submit host to the log file
=item	SUBMIT_STATUS the completion status of the globusrun_ws command
=item	STATUS the last status reported for the GRAM job

=back

=cut

package Job;

my $_jobindex = 0;

sub new
{
	my ( $type, $host, $timestamp, $logfile, $dbg ) = @_;
	# verbose prints data involved in intermediate steps
	# status prints everything that goes to the log file, and other status info
	$debug = $dbg;

	my @infiles = ();
	my $self = {
		HOST =>      $host,
		TIMESTAMP => $timestamp,
		LOGFILE => $logfile,
		_INDEX => \$_jobindex,	# corresponds to CPU compute core
		SUBMIT_STATUS => 0,
		STATUS =>  '',
		PREV_STATUS =>  '',
		PREV_STATUS_MSG =>  '',
		ACTIVE =>  0,
		FAILED =>  0,
		DONE =>  0,
		GRAM_SUBMIT_TIME => time(),
		SECS_LOCAL_SUBMIT =>      0,
		SECS_ACTIVE =>      0,
		INFILES => \@infiles
	};
	$_jobindex ++;
	my $this = bless $self, $type;
	$self->{INDEX} = ${ $self->{"_INDEX"} };
	return $this;
}

sub _log
{
	my $self = shift;
	my $msg = shift;
	my $idx = $self->getIndex();
	$self->{LOGFILE}->note( sprintf( "Job %4d: %s", $idx, $msg ) );
}

sub getIndex
{
	my $self = shift;
	return $self->{INDEX};
}

sub getEPRFile
{
	my $self = shift;
	return $tempfile_dir . $self->{TIMESTAMP} . "-"
		. $self->getIndex() . ".epr"
}

sub getHostname
{
	my $self = shift;
	return $self->{HOST}->name();
}

sub secs_since_gram_submit
{
	my $self = shift;
	return time() - $self->{GRAM_SUBMIT_TIME};
}

sub secs_since_active
{
	my $self = shift;
	return time() - $self->{SECS_ACTIVE};
}

my $joboutputs = '(?:Job failed:)'
        . '|(?:Error while loading job epr: globus_xio:)\s*(.+)(?:\n(.*))?';

sub updateStatus
{
	my $self = shift;

	if( ! $self->submit_status_good() )
	{
		return;
	}
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
		= localtime( time() );
	#FIXME would be better for everyting to get this info 
	# from a common source so it's formatted identically
	my $time = sprintf( "%02d-%02d-%02d_%02d.%02d.%02d",
		($year + 1900) % 100, $mon+1, $mday, $hour, $min, $sec );
	my $eprfile = $self->getEPRFile();
	my $cmd = "globusrun-ws -status -j $eprfile";
	my $globusrun_status = $? >> 8;
	my $output = `$cmd 2>&1`;
	chomp $output;
	#Current job state: StageIn

	$_ = $output;
	s/(?:Job failed:\s*(.*))?Current job state:\s*(.+)/$1$2/s;
	my $status = $2;
	my $statusmsg = $1;
	chomp $status;
	chomp $statusmsg;

	if( $status )
	{
		$self->{STATUS} = $status;
		if( $status ne $self->{PREV_STATUS}
			and ( grep( $debug, 'status' ) ) )
		{
			$self->_log( sprintf( "%17s Status %s",
							$time, $status ) );
		}
		if( $status eq 'Failed' )
		{
			$self->set_failed( 1 );
			$self->_log( sprintf( "%17s Failure message: %s",
						$time, $statusmsg ) );
			unlink $self->getEPRFile();
		}
		elsif( $status eq 'Done' )
		{
			$self->set_done( 1 );
			unlink $self->getEPRFile();
		}
	}
	else
	{
		s/$joboutputs/$1/m;
		my $status = $1;
		chomp $status;
		if( $status )
		{
			$self->{STATUS} = $status;
			$self->_log( sprintf( "%17s %s", $time, $status ) );
			$self->set_failed( 1 );
		}
		else
		{
			$status = 'unknown';
			$self->_log( sprintf( "%17s Error message: %s",
							$time, $output ) );
			$self->set_failed( 1 );
		}
	}
	$self->{PREV_STATUS} = $self->{STATUS};
	$self->{PREV_STATUS_MSG} = $output;
	# StageIn Pending StageOut CleanUp
}

sub getPrevStatus
{
	my $self = shift;
	return $self->{PREV_STATUS};
}

sub getPrevStatusMsg
{
	my $self = shift;
	return $self->{PREV_STATUS_MSG};
}

sub set_active
{
	my $self = shift;
	$self->{ACTIVE} = 1;
	$self->{SECS_ACTIVE} = time();
}

sub is_active
{
	my $self = shift;
	return $self->{ACTIVE} == 1;
}

sub is_done
{
	my $self = shift;
	return $self->{DONE} == 1;
}

sub failed
{
	my $self = shift;
	return $self->{FAILED} == 1;
}

sub set_done
{
	my $self = shift;
	$self->{DONE} = 1;
}

sub set_failed
{
	my $self = shift;
	$self->{FAILED} = 1;
}

sub toString
{
	my $self = shift;
	return __PACKAGE__ . ': ' . $self->{HOST} . ", " . $self->{STATUS}
		. $self->getEPRFile();
}

sub kill
{
	my $self = shift;
	my $eprfile = $self->getEPRFfile();
	my $host = $self->getHostname();
	if( -r $eprfile )
	{
		my $cmd = "globusrun-ws -kill -j $eprfile";
		my $output = `$cmd 2>&1`;
		unlink $eprfile;
		$self->_log( "killed on $host. Output: $output" );
	}
	else
	{
		$self->_log( "not killed on $host" .
			" -- EPR file '$eprfile' not found." );
	}
}

sub submit
{
	my $self = shift;

	my $inFilepaths = shift;
	my $outFilepaths = shift;

	my $hostName = $self->getHostname();
	my $timestamp = $self->{TIMESTAMP};
	my $num = $self->getIndex();

	$self->_log( "Trying host '$hostName'" );

	my $jobDir = "GDIST/$timestamp/$num";

	my $status = $self->make_cluster_dirs( $hostName, $jobDir );
	if( $status )
	{
		$self->{SUBMIT_STATUS} = $status;
		return;
	}

	push @{$self->{INFILES}}, @$inFilepaths;

	my @infilenames = $self->getInputFileNames();
	$self->_log( "Input files: " . join( ' ', @infilenames ) );

	my $jobDesc = $self->build_description( $jobDir, $outFilepaths );

	my $jobdesc_filename = $tempfile_dir . "job-$timestamp$num.desc.xml";

	$self->write_descripion_file( $jobdesc_filename, $jobDesc );

	my $run_cmd = $self->get_run_command_line( $jobdesc_filename );

	my $run_output = `$run_cmd 2>&1`;

	$self->{SUBMIT_STATUS} = $?;

	if( $self->{SUBMIT_STATUS} )
	{
		chomp $run_output;
		$self->_log( "GRAM submission failed: \n\t$run_output\n" );
	}
	else
	{
		$self->_log( "GRAM submission succeeded." );
	}

	unlink $jobdesc_filename;
}
=pod

=head2 build_description()

=over 1

Based on the input and output file names, builds a Globus GRAM
job description in the JDD (Job Description Document) language.
L<http://www.globus.org/toolkit/docs/4.0/execution/wsgram/schemas/gram_job_description.html>

The choice of the JDD language is for Globus 4.0.x support.  If only later
versions of Globus are to be used, it might be better to use JSDL.

The F<mult-proc.pl> file is set to be staged in and executed.  It is this
file that launches multiple processes based on the number of CPU cores.

=back

=head3 Requires

=over 2

=item	a directory F<$HOME/GDIST/>

=item	a script F<$HOME/GDIST/multi-proc.pl>

=back

=head3 Parameters

=over 2

=item	the path on the compute host to a working directory (relative to
        the $GLOBUS_USER_HOME)

=item	a reference to a list of paths to the input files on the host
        (either full or relative)

=item	the path on the host to the output directory (either full or relative)

=back

=cut

sub build_description
{
	my $self = shift;
	my $workdir = shift;
	my $out_dirpath = shift;

	my $host = $self->{HOST};

	my $submit_host = Net::Domain::hostfqdn();
	my $pwd;
	chomp( $pwd = `pwd` );
	#$pwd = substr( $pwd, 1 ); # remove leading slash

	my $queue = $host->queueName();

	my @in_filenames = $self->getInputFileNames();

	my $jdd = '';
	my $xw = new XML::Writer( OUTPUT => \$jdd,
			DATA_MODE => 1, DATA_INDENT => 1 );

	#FIXME would it be appropriate to do this in GLOBUS_SCRATCH_DIR?
	# would really prefer to use local disks... 
	# but is that what GLOBUS_SCRATCH_DIR is for?
	my $jobdir = '${GLOBUS_USER_HOME}/' . $workdir;
	my $submitdir = "$submit_host$pwd";

	$xw->startTag( 'job' );
	$xw->dataElement( 'executable', './multi-proc.pl' );
	$xw->dataElement( 'directory', "$jobdir/" );
	for my $in_filename (@in_filenames)
	{
		$xw->dataElement( 'argument', $in_filename );
	}
	$xw->dataElement( 'stdout', "$jobdir/stdout" );
	$xw->dataElement( 'stderr', "$jobdir/stderr" );
	$xw->dataElement( 'count', '1' );
	if( $queue )
	{
		$xw->dataElement( 'project', 'GDIST' );
		$xw->dataElement( 'queue', $queue );
		$xw->dataElement( 'jobType', 'single' );
#		$xw->startTag( 'extensions' );
#			$xw->dataElement( 'jobname', 'GDIST' );
#		$xw->endTag( 'extensions' );
	}
	$xw->startTag( 'fileStageIn' );
	$xw->startTag( 'transfer' );
	$xw->dataElement( 'sourceUrl', "gsiftp://$submitdir/multi-proc.pl" );
	$xw->dataElement( 'destinationUrl', "file:///$jobdir/" );
	$xw->endTag( 'transfer' );
	$xw->startTag( 'transfer' );
	$xw->dataElement( 'sourceUrl', "gsiftp://$submitdir/single.sh" );
	$xw->dataElement( 'destinationUrl', "file:///$jobdir/" );
	$xw->endTag( 'transfer' );

	my $i = 0;
	for my $in_filepath (@{$self->{INFILES}})
	{
		$i++;
		$xw->startTag( 'transfer' );
		my ($vol, $dir, $in_filename)
			= File::Spec->splitpath( $in_filepath );
		$xw->dataElement( 'sourceUrl', "gsiftp://$submit_host$in_filepath" );
		$xw->dataElement( 'destinationUrl', "file:///$jobdir/$in_filename" );
		$xw->endTag( 'transfer' );
	}

	$xw->endTag( 'fileStageIn' );

	$xw->startTag( 'fileStageOut' );
	$i = 0;
	for my $in_filename (@in_filenames)
	{
		$i++;
		$xw->startTag( 'transfer' );
		$xw->dataElement( 'sourceUrl', "file:///$jobdir/$in_filename" );
		$xw->dataElement( 'destinationUrl', "gsiftp://$submit_host$out_dirpath/$in_filename" );
		$xw->endTag( 'transfer' );
	}
	$xw->endTag( 'fileStageOut' );

	$xw->startTag( 'fileCleanUp' );
	$i = 0;
	for my $in_filename (@in_filenames)
	{
		$i++;
		$xw->startTag( 'deletion' );
		$xw->dataElement( 'file', "file:///$jobdir/$in_filename" );
		$xw->endTag( 'deletion' );
	}
	$xw->startTag( 'deletion' );
	$xw->dataElement( 'file', "file:///$jobdir/multi-proc.pl" );
	$xw->endTag( 'deletion' );
	$xw->startTag( 'deletion' );
	$xw->dataElement( 'file', "file:///$jobdir/single.sh" );
	$xw->endTag( 'deletion' );
	#$xw->startTag( 'deletion' );
	#$xw->dataElement( 'file', "file:///$jobdir/" );
	#$xw->endTag( 'deletion' );
	$xw->endTag( 'fileCleanUp' );

	$xw->startTag( 'extensions' );
		$xw->dataElement( 'emailontermination', 'no' );
		$xw->dataElement( 'emailonexecution', 'no' );
		$xw->dataElement( 'emailonsuspend', 'no' );
		$xw->dataElement( 'emailonabort', 'no' );
	$xw->endTag( 'extensions' );

	$xw->endTag( 'job' );

	$xw->end();

	( grep /verbose/, $debug ) and print $jdd;

	return $jdd;
}

=pod

=head2 getInputFileNames()

=over 1

returns the list of names of input files, stripped of file path info

=back

=cut

sub getInputFileNames
{
	my $self = shift;
	my @in_filepaths = @{$self->{INFILES}};
	my @in_filenames = ();
	for my $in_filepath (@in_filepaths)
	{
		my ($vol, $dir, $name) = File::Spec->splitpath( $in_filepath );
		push @in_filenames, $name;
	}
	return @in_filenames;
}

=pod

=head2 get_run_command_line()

=over 1

Constructs the command line to start the C<globurun-ws> command.

=back

=cut

sub get_run_command_line
{
	my $self = shift;
	my $host = $self->{HOST};
	my $jobdesc_name = shift;

	my $scheduler = '';
	if( $host->scheduler() )
	{
		$scheduler = "-Ft " . $host->scheduler() . " ";
	}

	my $run_cmd = 'globusrun-ws -q -submit -b -S ' . $scheduler
			. " -T $ws_timeout_msec"
			. ' -F ' . $self->getHostname() . " -f $jobdesc_name"
			. ' -o ' . $self->getEPRFile();

	( grep /status/, $debug ) and print "Run command: $run_cmd\n";

	return $run_cmd;
}

=pod

=head2 make_cluster_dirs()

=over 1

Make a unique dir for this job on cluster

=back

=cut

sub make_cluster_dirs
{
	my $self = shift;
	my $hostname = shift;
	my $jobdir_name = shift;

	my $mkdir_cmd = "mkdir -p $jobdir_name";
	my $gsissh_cmd = "gsissh $hostname \'$mkdir_cmd\'";
	my $shell_cmd = "bash -c \"$gsissh_cmd\" 2>&1";
	my $gsissh_output = `$shell_cmd`;

	my $status = $? >> 8;
	if( $status )
	{
		chomp $gsissh_output;
		my $msg = "Could not do gsissh log-in to $hostname\n"
			. "gsissh error $status\n     $gsissh_output\n\n";
		$self->_log( $msg );
	}
	return $status;
}

=pod

=head2 write_descripion_file()

=over 1

Write a file in current directory, with given content. 
Die on failure with error message.

=back

=cut

sub write_descripion_file
{
	my $self = shift;
	my $filename = shift;
	my $job_description = shift;

	my $file;
	open $file, ">$filename" 
		or die "Could not write job description file: $filename: $!";
	print $file $job_description;
	close $file;
}

=pod

=head2 write_EPR_file()

=over 1

Write a file in current directory, with given content. 
Die on failure with error message.

=back

=cut

sub write_EPR_file
{
	my $self = shift;
	my $epr = shift;
	my $filename = $self->getEPRFile();

	my $file;
	open $file, ">$filename" 
		or die "Could not write job EPR file: $filename: $!";
	print $file $epr;
	close $file;
}

=pod

=head2 submit_status_good()

=over 1

Returns true if the status of the grid job terminated successfully.

=back

=cut

sub submit_status_good
{
	my $self = shift;
	return $self->{SUBMIT_STATUS} == 0;
}

1;
