#!/usr/bin/perl
use strict;

use XML::Writer;
use Net::Domain qw(hostfqdn);	# $ENV{HOST} and system hostname are unreliable
use File::Spec;
use List::Util qw[min max];
use Getopt::Long;

use Job;
use LogFile;
use GridHost;


=head1 NAME

grid_dist

=head1 DESCRIPTION

Runs multiple compute processes on a grid cluster resource.

The user account should correspond to a grid user whose grid certificates
are installed, and a valid proxy should be present.

One Globus job is run per node, which will run n processes on that node
if it has n compute cores.

=head1 AUTHOR

Steve White <swhite@aip.de>

=head1 VERSION

$Revision: 278 $
$Date: 2010-02-08 17:36:18 +0100 (Mon, 08 Feb 2010) $

=head1 FUNCTIONS

=cut

my $debug = '';
# verbose prints data involved in intermediate steps
# status prints everything that goes to the log file, and other status info

my $vrb = '';
my $sta = '';
my $ncores_lim = '';
GetOptions( 'verbose' => \$vrb,
'status' => \$sta,
'limit-cores=i' => \$ncores_lim );

if( $vrb )
{
	$debug .= ' verbose';
}
if( $sta )
{
	$debug .= ' status';
}

my $tempfile_dir = $ENV{HOME} . '/.grid_dist/';

my $ws_timeout_msec = 3 * 60 * 1000;

if( @ARGV < 2 )
{
	die( "Usage: \n"
	   . "\tperl grid_dist.pl out_dirpath in_filepath1 in_filepath2... .\n" 
	   . "Flags:\n"
	   . "\t--verbose\n"
	   . "\t--status\n"
	   . "\t--limit-cores=n\n" );
}

my $out_dirpath = File::Spec->rel2abs( @ARGV[0] );

my @infiles = ();

for my $f ( @ARGV[1..$#ARGV] )
{
	 push( @infiles, File::Spec->rel2abs( $f ) );
}


my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime( time() );
my $timestamp = sprintf( "%02d-%02d-%02d_%02d.%02d.%02d_%07d",
		($year + 1900) % 100, $mon + 1, $mday, $hour, $min, $sec, $$ );

my $log = new LogFile( $timestamp, $tempfile_dir, $debug );

$log->note( "Output dir argument: $out_dirpath\n" );
$log->note( 'Input file arguments: ' . join( ' ', @infiles ) . "\n" );

check_certificate();
check_IO_paths( \@infiles, $out_dirpath );

my @files_left = @infiles;

my @hosts = get_available_hosts();

my %jobs = ();

# FIXME the logic of this loop is all wrong.
# currently it just cycles through hosts in case the job isn't submitted.
# but often a submitted job fails.  In this case, the script just 
# awaits a termination that never happens.
while( @files_left )
{
	if( @hosts )
	{
		my $host = $hosts[0];
		my $ncores = $host->numberOfCores();
		if( $ncores_lim )
		{
			$ncores = $ncores_lim;
		}
		elsif ( $ncores < 1 )
		{
			$ncores = 1;
		}
		my $nfiles_to_submit = min( $ncores, $#files_left + 1 );
		my @in_files = @files_left[0..$nfiles_to_submit - 1];

		my $job = new Job( $host, $timestamp, $log, $debug );
		$job->submit( \@in_files, $out_dirpath );

		if( $job->submit_status_good() )
		{
			$jobs{$job->getIndex()} = $job;
			@files_left = @files_left[$ncores..$#files_left];
			sleep 5;	# by request of S. Freitag
		}
		else
		{
			shift @hosts;
		}
	}
	else
	{
		exit( 1 );
	}
	# FIXME check that files have been submitted
}

await_termination( \%jobs );

exit( 0 );

=pod

=head2 check_certificate()

=over 1

Dies if the current user's certificate
	* is not found
	* is not valid
	* does not have a valid proxy

=back

=cut

sub check_certificate
{
	my $certcheck = "grid-proxy-info|grep timeleft";
	open( CH, "$certcheck|" ) or die( "Problem checking grid certificate" );
	while( <CH> )
	{
		if( / 0\:00\:00/ )
		{
			die( "No time left on grid proxy.  "
				. "Run grid-proxy-init.\n" );
		}
	}
	close( CH );
}

=pod

=head2 get_available_hosts()

=over 1

Gets a list of GridHost objects.  

In the absence of a functioning grid scheduler, it just takes a static list.

=back

=cut

sub get_available_hosts
{
	#FIXME... this probably looks in current dir-- should be script dir
	my $hosts_file = "grid_cluster_hosts.csv";
	my $hfh;
	open $hfh, "<$hosts_file"
		or die( "Couldn't read hosts file $hosts_file." );
	my @hosts = ();
	while( <$hfh> )
	{
		my @tokens = split( '#' );
		if( @tokens )
		{
			my $before_hash = $tokens[0];
			my @host_args = split( /\s*,\s*/, $before_hash );
			chomp @host_args;
			if( $host_args[0] )
			{
				push @hosts, new GridHost( \@host_args );
			}
		}
	}

	if( grep /verbose/, $debug )
	{
		for my $host (@hosts )
		{
			print $host->toString( ) . "\n";
		}
	}
	return @hosts;
}

=pod

=head2 await_termination()

=over 1

Takes a reference to a list of Jobs.

Copies this list, and works with the copy.

Iterates through the copy of the list list, calling updateStatus() on each,
then calling trim_non_running_jobs() to remove any dead, done, or
timed-out jobs from the list.

=back

=cut

sub await_termination
{
	my $jobs = shift;
	my %jobs = %$jobs;
	my %remaining_jobs = ();

	for my $i (keys %jobs )
	{
		$remaining_jobs{$i} = $jobs{$i};
	}

	while( %remaining_jobs )
	{
		( grep /verbose/, $debug ) and print "Jobs yet to finish: "
					. keys( %remaining_jobs ) . "\n";

		for my $i (keys %remaining_jobs)
		{
			my $js = $remaining_jobs{$i};
			$js->updateStatus();
			trim_non_running_jobs( \%remaining_jobs, $i );
		}
		if( %remaining_jobs )
		{
			sleep 10;
		}
	}
	#FIXME what to do about jobs that aren't finished?
}

=pod

=head2 trim_non_running_jobs()

=over 1

Takes a reference to a hash of Job, and a key for a certain Job in the hash.

If the job has failed or is done, removes the job from the hash.

If the job has timed out, kills the job and removes it from the hash.

=back

=cut

sub trim_non_running_jobs
{
	my $remaining_jobs = shift;
	my $i = shift;
	my $js = $remaining_jobs->{$i};
	my $secs_for_active = 3 * 60 * 60;	#  3 hours
	my $secs_for_success = 12 * 60 * 60;	# 12 hours

	#( grep /verbose/, $debug ) and print "Job $i $output\n";
	# FIXME this is very ugly and certainly wrong
	if( $js->failed() or $js->is_done() )
	{
		delete $remaining_jobs->{$i};
	}
	else
	{
		if( $js->is_active() )
		{
			my $t = $js->secs_since_active();
			if ( $t > $secs_for_success )
			{
				$log->note( "Job $i timed out "
					. "while active." );
				$js->kill();
				delete $remaining_jobs->{$i};
			}
		}
	}
	if( ! $js->is_active() )
	{
		if( 0 )	#FIXME how to do this right?
		{
			my $t = $js->secs_since_gram_submit();
			if ( $t > $secs_for_active )
			{
				$log->note( "Job $i timed out "
					. "before activated." );
				$js->kill();
				delete $remaining_jobs->{$i};
			}
		}
	}
}

=pod

=head2 check_IO_paths()

=over 1

Dies if the input file does not exist or does not appear to be an archive
or isn't readable by the grid user.

Dies if the output file path is not writable by the grid user.

=back

=cut

sub check_IO_paths
{
	my $infile_ref = shift;
	my @infiles = @$infile_ref;
	my $out_dir = shift;
	#$out_dir =~ s/(.*)\/.*/$1/;

	my $hostname = hostfqdn();

	for my $f ( @infiles )
	{
		if( not $f )
		{	
			die "No input file '$f' found.";
		}

		if( not grep /.*\.tar.gz|.*\.tgz|.*\.zip/, $f )
		{
			die "Input file doesn't appear to be an archive: $f";
		}
		my $test = "[ ! -r $f ] && echo 1";
		my $ret = `gsissh $hostname "bash -c '$test'"`;

		if( $ret =~ /^1$/ )
		{
			die "Input file not readable by grid user $f";
		}
	}

	my $test = "( [ ! -d $out_dir ] || [ ! -w $out_dir ] ) && echo 2";
	my $ret = `gsissh $hostname 'bash -c "$test"'`;

	if( $ret =~ /^2$/ )
	{
		die "Output directory not writable by grid user $out_dir";
	}
}

