#!/usr/bin/perl
use strict;

use File::Path;

=pod

=head1 NAME

multi_proc.pl

=head1 DESCRIPTION

 Forks a child process for each argument, feeding each to a script
	single.sh

 Intended for use on multi-core machines, where each process will be
 run on a dedicated core.

 Each child runs in a sequentially-numbered directory created by the parent.

 The parent waits for all child processes to exit, and checks the status.

=head1 USE

multi_proc.pl file1 file2...

=head1 AUTHOR

Steve White <swhite@aip.de>

=head1 VERSION

$Revision: 197 $
$Date: 2009-12-21 17:59:59 +0100 (Mon, 21 Dec 2009) $

=cut

( @ARGV ) or die "No arguments provided.\n";

print "Executing processes on $ENV{HOSTNAME}\n";

my %kids = ();

my $jobdir = `pwd`;
chomp $jobdir;

my $tmpdir = '';

if( exists $ENV{TMPDIR} )
{
	$tmpdir = "$ENV{TMPDIR}/$$/";
}

my $procno = 0;
for my $arg (@ARGV)
{
	$procno++;

	# Each process gets a separate directory for its I/O files,
	# and if $TMPDIR is defined, it also gets a separate direcotory
	# for the temporary files.
	# $workpath will typically be in a separate filesystem from $procpath.
	my $workpath = "$tmpdir$procno";

	mkpath( "$workpath" ) or die "Could not make dir $workpath";
	
	my $child = fork();

	if( $child == 0 )	# done only by child processes
	{
		chdir "$workpath" or die "Could not chdir $workpath";
		print "EXEC'ING single.sh '$arg' '$jobdir' in $workpath\n";

		my $cmd = "$jobdir\/single.sh '$arg' '$jobdir'";
		exec( $cmd ); # child becomes cmd; doesn't return
	}
	$kids{$child} = $procno;
}
use POSIX ":sys_wait_h";
while( %kids )
{
	my $pid = waitpid( -1, WNOHANG );
	if( $pid > 0 )
	{
		my $num = $kids{$pid};
		if( $? )
		{
			print STDERR "Process $num with pid $pid exited with "
					. "status $?.\n"
		}
		else
		{
			print STDOUT "Process $num with pid $pid completed.\n";
		}
		delete $kids{$pid};
		my $wd = "$tmpdir$num";
		rmtree( "$wd" ) or print STDERR "Could not delete dir $wd/";
	}
	sleep( 1 )
};
print "All processes exited.\n";
